#include "MeshLoader.h"
#include <set>
#include <list>
#include <unordered_set>
#include <array>
#include <algorithm>
#include <ranges>
#include<string>
#include <fstream>
MeshLoader::~MeshLoader()
{
}

std::vector<Node>& MeshLoader::GetNodes() {
	return nodes;
}
Node MeshLoader::get_node(rsize_t id) {
    return nodes.at(id);
}


std::vector<FiniteElement>& MeshLoader::GetFE() {
	return finiteElements;
}

std::vector<BoundaryFiniteElement>& MeshLoader::GetBFE() {
	return boundaryFiniteElements;
}

void MeshLoader::Print_Data()
{
    std::cout << &GetNodes() << std::endl
        << &GetFE() << std::endl
        << &GetBFE() << std::endl;
}

//!!! ИСПРАВЛЕНО Контейнеры-результаты должны возвращать ID, а не сами узлы и КЭ

std::vector<int> MeshLoader::findFE(size_t ID1, size_t ID2, size_t ID3) {
    std::vector<int> ans;
    auto cur_iter = finiteElements.begin();
    auto last_iter = finiteElements.end();

    while (cur_iter != last_iter) {
        cur_iter = std::find_if(cur_iter, last_iter, [ID1, ID2, ID3](FiniteElement& cur_elem) {
            return std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID1) != cur_elem.materials.end()
                && std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID2) != cur_elem.materials.end()
                && std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID3) != cur_elem.materials.end();
            });

        if (cur_iter != last_iter) {
            ans.push_back(cur_iter->ID);
            ++cur_iter;
        }
    }

    return ans;

}


std::vector<FiniteElement> MeshLoader::findEdge(size_t ID1, size_t ID2) {
    std::vector<FiniteElement> ans;
    auto cur_iter = finiteElements.begin();
    auto last_iter = finiteElements.end();

    while (cur_iter != last_iter) {
        cur_iter = std::find_if(cur_iter, last_iter, [ID1, ID2](FiniteElement& cur_elem) {
            return std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID1) != cur_elem.materials.end()
                && std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID2) != cur_elem.materials.end();
            });

        if (cur_iter != last_iter) {
            ans.push_back(*cur_iter);
            ++cur_iter;
        }
    }

    return ans;

}

std::vector<Node> MeshLoader::vertex_nodes_by_surface_ID(size_t s_id) { //!!! ИСПРАВЛЕНО Не корректное название метода. Что за ID поверхзности?

        std::vector<Node> answer;
        auto current_iter = boundaryFiniteElements.begin();
        while (current_iter != boundaryFiniteElements.end()) {
            current_iter = std::find_if(current_iter, boundaryFiniteElements.end(), [s_id](const BoundaryFiniteElement& s) { return s.ID_surface == s_id; });
            if (current_iter != boundaryFiniteElements.end())
            {
                for (const auto& elem : current_iter->node_boundary) { //!!!ИСПРАВЛЕНО Неверное имя: it - это не итератор
                    
                    //!!! ИСПРАВЛЕНО Крайне не эффективно. Потеряны преимущества контейнера узлов.
                    
                    answer.insert(Get_Node_by_ID(s_id)); //!!! ИСПРАВЛЕНО Где гарантия, что не будет дубликатов? Разные КЭ могут содержать одни и те же узлы.
                }
                ++current_iter;
            }
        }
        return answer;
    }


std::vector<Node> MeshLoader::getMaterialID(size_t ID) {
    std::list<Node> result;
    auto element_by_material = nodes.begin();
    auto last_element = nodes.end();
    while (element_by_material != last_element) {
        
        element_by_material = std::find_if(element_by_material, last_element, [ID](const Node& e) { return e.ID_node == ID; });//!!! ИСПРАВЛЕНО Почему e по значению?
        
        if (element_by_material != last_element) {
            result.push_back(*element_by_material);
            ++element_by_material;
        }
    }
    return std::vector<Node>(result.begin(), result.end());
}

std::vector<std::set<size_t>>  MeshLoader::GetNodesNeighbors()
{
    std::vector<std::set<size_t>> result(nodes.size());
    for (const auto& it_FiniteElement : boundaryFiniteElements)
    {
        for (auto& it_nodes : it_FiniteElement.node_boundary)
        {
            result[it_nodes - 1].insert(it_FiniteElement.node_boundary.begin(), it_FiniteElement.node_boundary.end());
            result[it_nodes - 1].erase(it_nodes);
        }
    }

    return result;
}

std::vector<BoundaryFiniteElement> MeshLoader::getBoundaryFiniteID(size_t ID) {
    std::list<BoundaryFiniteElement> result;
    auto surface_by_ID = boundaryFiniteElements.begin();
    auto last_element = boundaryFiniteElements.end();
    while (surface_by_ID != last_element) {
        surface_by_ID = std::find_if(surface_by_ID, last_element, [ID](const BoundaryFiniteElement& s) { return s.ID_surface == ID; }); //!!! ИСПРАВЛЕНО Почему s по значени?
        if (surface_by_ID != last_element) {
            result.push_back(*surface_by_ID);
            ++surface_by_ID;
        }
    }
    return std::vector<BoundaryFiniteElement>(result.begin(), result.end());
}

Node MeshLoader::Get_Node_by_ID(size_t n_id){
    
    //!!! ИСПРАВЛЕНО Крайне не эффективно. При известном ID просто так обходить все узлы?
    return nodes.at(n_id);
}

void MeshLoader::InsertNewNodes(){
    std::unordered_set<Edge, EdgeHash> edge_container;
    for (auto& elem : boundaryFiniteElements) //!! ИСПРАВЛЕНО Где вы тут итератор нашли? Давайте осмысленные имена
    {
        std::vector<int> current_iter(elem.ID_boundary);
        for (auto i = current_iter.begin(); i != current_iter.end(); i++) //!!!ИСПРАВЛЕНО -1 откуда взялось?
        {
            for (auto j = i + 1; j != current_iter.end(); j++)
            {
                Edge current_edge(*i, *j);
                auto search_edge = edge_container.find(current_edge);

                if (search_edge == edge_container.end()) {
                    Node left_node = Get_Node_by_ID(*i); //!!! ИСПРАВЛЕНО Копирование узла на каждой итерации - плохая иделя
                    Node right_node = Get_Node_by_ID(*j);

                    //!!! ИСПРАВЛЕНО Т.е. если search_edge != edge_container.end(), то создавали зря...
                    std::vector<float> x = { (left_node.coord[0] + right_node.coord[0]) / 2, (left_node.coord[0] + right_node.coord[0]) / 2, (left_node.coord[0] + right_node.coord[0]) / 2 };
                    Node center_node(nodes.size() + 1, x, false);
                    current_edge.self_id = nodes.size() + 1;
                    edge_container.insert(current_edge);
                    nodes.push_back(center_node);
                    elem.node_boundary.emplace_back(current_edge.self_id);
                }
                else{
                    elem.node_boundary.emplace_back(search_edge->self_id);
                }

            }
        }
    }

    for (auto& it : boundaryFiniteElements)
    {
        std::vector<size_t> current_iter(it.node_boundary);
        for (auto i = current_iter.begin(); i != current_iter.end() - 1; i++)
        {
            for (auto j = i + 1; j != current_iter.end(); j++)
            {
                Edge search_edge(*i, *j);
                auto search_point = edge_container.find(search_edge);
                if (search_point != edge_container.end())
                {
                    it.node_boundary.emplace_back(search_point->self_id);
                }
            }
        }
    }


}

bool MeshLoader::ratio( size_t id1, size_t id2, size_t id3, size_t id4) {
    Node A = get_node(id1);
    Node B = get_node(id2);
    Node C = get_node(id3);
    Node D = get_node(id4);
    auto len = [](Node x, Node y) {
        return sqrt((x.coord[0] - y.coord[0]) * (x.coord[0] - y.coord[0]) 
                  + (x.coord[1] - y.coord[1]) * (x.coord[1] - y.coord[1]) 
                  + (x.coord[2] - y.coord[2]) * (x.coord[2] - y.coord[2]));
    };
    float A1 = std::max(std::max(std::max(len(A,B), len(A, C)), std::max(len(A, D), len(B, C))), std::max(len(B, D), len(C, D)));
    float A2 = (0.5 * (3 * A1)) * A1 + (sqrt(3) / (4 * (A1 * A1)));
    return A2 / A1>1;
}
struct tetr {
    int id1, id2, id3, id4;

    tetr() : id1(0), id2(0), id3(0), id4(0) {}
    tetr(int ID1, int ID2, int ID3, int ID4) : id1(ID1), id2(ID2), id3(ID3), id4(ID4) {}

    friend std::ostream& operator << (std::ostream& out, const tetr& ABCD) {
        out << ABCD.id1 << " " << ABCD.id2 << " " << ABCD.id3 << " " << ABCD.id4 << std::endl;
        return out;
    }
};

void MeshLoader::sort_range(const std::string& filepath) {

    std::vector<FiniteElement> elements = GetFE();
    if (elements.empty())
        return;

    std::vector<tetr> tetrahedrons;
    std::ofstream file(filepath + ".txt");
    for (int i : std::ranges::views::iota(size_t(0), elements.size())) {
        for (int j : std::ranges::views::iota(size_t(0), elements.size())) {
            for (int k : std::ranges::views::iota(size_t(0), elements.size())) {
                for (int l : std::ranges::views::iota(size_t(0), elements.size()))
                    tetrahedrons.push_back(tetr(elements[i].materials[0], elements[j].materials[0], elements[k].materials[0], elements[l].materials[0]));
            }
        }
    }

    std::ranges::copy(tetrahedrons | std::ranges::views::filter([&](const tetr& ABCD) {
        return ratio(ABCD.id1, ABCD.id2, ABCD.id3, ABCD.id4);
        }
    ) | std::ranges::views::transform(
            [](const tetr& ABCD) {
                return ABCD;
            }), std::ostream_iterator<tetr>(file));

}
